-------------------------------------------------------------------------------
--! @file       CipherCore.vhd
--! @author     Ekawat (ice) Homsirikamol
--! @brief      CipherCore for Deoxys and Joltik
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.ALL;
use work.AEAD_pkg.all;

entity CipherCore is
    generic (
        --! Reset behavior
        G_ASYNC_RSTN    : boolean := False; --! Async active low reset
        --! Block size (bits)
        G_DBLK_SIZE     : integer := 128;   --! Data
        G_KEY_SIZE      : integer := 32;    --! Key
        G_TAG_SIZE      : integer := 128;   --! Tag
        --! The number of bits required to hold block size expressed in
        --! bytes = log2_ceil(G_DBLK_SIZE/8)
        G_LBS_BYTES      : integer := 4;
        --! Maximum supported AD/message/ciphertext length = 2^G_MAX_LEN-1
        G_MAX_LEN       : integer := SINGLE_PASS_MAX
    );
    port (
        --! Global
        clk             : in  std_logic;
        rst             : in  std_logic;
        --! PreProcessor (data)
        key             : in  std_logic_vector(G_KEY_SIZE       -1 downto 0);
        bdi             : in  std_logic_vector(G_DBLK_SIZE      -1 downto 0);
        --! PreProcessor (controls)
        key_ready       : out std_logic;
        key_valid       : in  std_logic;
        key_update      : in  std_logic;
        decrypt         : in  std_logic;
        bdi_ready       : out std_logic;
        bdi_valid       : in  std_logic;
        bdi_type        : in  std_logic_vector(3                -1 downto 0);
        bdi_partial     : in  std_logic;
        bdi_eot         : in  std_logic;
        bdi_eoi         : in  std_logic;
        bdi_size        : in  std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
        bdi_valid_bytes : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
        bdi_pad_loc     : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
        --! PostProcessor
        bdo             : out std_logic_vector(G_DBLK_SIZE      -1 downto 0);
        bdo_valid       : out std_logic;
        bdo_ready       : in  std_logic;
        bdo_size        : out std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
        msg_auth_done   : out std_logic;
        msg_auth_valid  : out std_logic
    );
end entity CipherCore;

architecture structure of CipherCore is
    constant ADDR_WIDTH   : integer := 3+128/G_DBLK_SIZE;
    signal addr           : std_logic_vector(ADDR_WIDTH           -1 downto 0);
    signal stage_status   : std_logic_vector(4                    -1 downto 0);
    signal ld_block       : std_logic;
    signal ld_tweakey     : std_logic;
    signal ld_chksum      : std_logic;
    signal ld_auth        : std_logic;
    signal ld_state       : std_logic;
    signal en_auth        : std_logic;
    signal en_block       : std_logic;
    signal en_tweakey     : std_logic;
    signal en_chksum      : std_logic;
    signal en_state       : std_logic;
    signal en_key         : std_logic;
    signal en_npub        : std_logic;
    signal sel_input      : std_logic_vector(2                    -1 downto 0);
    signal sel_chksum     : std_logic;
    signal sel_rdkey      : std_logic;
    signal sel_output     : std_logic;
    signal sel_decrypt    : std_logic;
    signal sel_npub       : std_logic;
    signal sel_tag        : std_logic;
begin

    u_cc_dp:
    entity work.CipherCore_Datapath(structure)
    generic map (G_MAX_LEN => G_MAX_LEN)
    port map (
        clk                 => clk              ,
        rst                 => rst              ,

        --! Input Processor
        key                 => key              ,
        bdi                 => bdi              ,
        bdi_valid_bytes     => bdi_valid_bytes  ,
        bdi_pad_loc         => bdi_pad_loc      ,

        --! Output Processor
        bdo                 => bdo              ,
        msg_auth_valid      => msg_auth_valid   ,

        --! Controller
        addr                => addr             ,
        stage_status        => stage_status     ,
        ld_block            => ld_block         ,
        ld_tweakey          => ld_tweakey       ,
        ld_chksum           => ld_chksum        ,        
        ld_auth             => ld_auth          ,
        ld_state            => ld_state         ,
        en_auth             => en_auth          ,
        en_block            => en_block         ,        
        en_tweakey          => en_tweakey       ,        
        en_chksum           => en_chksum        ,
        en_state            => en_state         ,
        en_key              => en_key           ,
        en_npub             => en_npub          ,
        sel_input           => sel_input        ,
        sel_chksum          => sel_chksum       ,
        sel_rdkey           => sel_rdkey        ,
        sel_output          => sel_output       ,
        sel_decrypt         => sel_decrypt      ,
        sel_npub            => sel_npub         ,
        sel_tag             => sel_tag          
    );

    u_cc_ctrl:
    entity work.CipherCore_Control(behavior)
    generic map (
        G_ADDR_WIDTH        => ADDR_WIDTH       ,
        G_LBS_BYTES         => G_LBS_BYTES
    )
    port map (
        clk                 => clk              ,
        rst                 => rst              ,

        --! Input
        key_ready           => key_ready        ,
        key_valid           => key_valid        ,
        key_update          => key_update       ,
        decrypt             => decrypt          ,
        bdi_ready           => bdi_ready        ,
        bdi_valid           => bdi_valid        ,
        bdi_type            => bdi_type         ,
        bdi_eot             => bdi_eot          ,
        bdi_eoi             => bdi_eoi          ,
        bdi_size            => bdi_size         ,

        --! Datapath
        addr                => addr             ,
        stage_status        => stage_status     ,
        ld_block            => ld_block         ,
        ld_tweakey          => ld_tweakey       ,
        ld_chksum           => ld_chksum        ,        
        ld_auth             => ld_auth          ,
        ld_state            => ld_state         ,
        en_auth             => en_auth          ,
        en_block            => en_block         ,        
        en_tweakey          => en_tweakey       ,        
        en_chksum           => en_chksum        ,
        en_state            => en_state         ,
        en_key              => en_key           ,
        en_npub             => en_npub          ,
        sel_input           => sel_input        ,
        sel_chksum          => sel_chksum       ,
        sel_rdkey           => sel_rdkey        ,
        sel_output          => sel_output       ,
        sel_decrypt         => sel_decrypt      ,
        sel_npub            => sel_npub         ,
        sel_tag             => sel_tag          ,

        --! Output
        msg_auth_done       => msg_auth_done    ,
        bdo_ready           => bdo_ready        ,
        bdo_valid           => bdo_valid
    );

end structure;