-------------------------------------------------------------------------------
--! @file       CipherCore_Control.vhd
--! @author     Ekawat (ice) Homsirikamol
--! @brief      Control unit for Deoxys and Joltik
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.AEAD_pkg.all;

entity CipherCore_Control is
    generic (
        G_ADDR_WIDTH     : integer := 4;
        G_LBS_BYTES      : integer := 4
    );
    port (
        clk              : in  std_logic;
        rst              : in  std_logic;

        --! Input
        key_ready        : out std_logic;
        key_valid        : in  std_logic;
        key_update       : in  std_logic;
        decrypt          : in  std_logic;
        bdi_ready        : out std_logic;
        bdi_valid        : in  std_logic;
        bdi_type         : in  std_logic_vector(3               -1 downto 0);
        bdi_eot          : in  std_logic;
        bdi_eoi          : in  std_logic;
        bdi_size         : in  std_logic_vector(G_LBS_BYTES+1   -1 downto 0);

        --! Datapath
        addr             : out std_logic_vector(G_ADDR_WIDTH    -1 downto 0);
        stage_status     : out std_logic_vector(4               -1 downto 0);
        ld_block         : out std_logic;       
        ld_tweakey       : out std_logic;       
        ld_chksum        : out std_logic;       
        ld_auth          : out std_logic;       
        ld_state         : out std_logic;       
        en_auth          : out std_logic;       
        en_block         : out std_logic;       
        en_tweakey       : out std_logic;       
        en_chksum        : out std_logic;       
        en_state         : out std_logic;       
        en_key           : out std_logic;       
        en_npub          : out std_logic;       
        sel_input        : out std_logic_vector(2               -1 downto 0);
        sel_chksum       : out std_logic;
        sel_rdkey        : out std_logic;
        sel_output       : out std_logic;
        sel_decrypt      : out std_logic;
        sel_npub         : out std_logic;
        sel_tag          : out std_logic;

        --! Output
        msg_auth_done    : out std_logic;
        bdo_ready        : in  std_logic;
        bdo_valid        : out std_logic
    );
end entity CipherCore_Control;

architecture behavior of CipherCore_Control is    
    --! Determine the round size
    impure function get_round return integer is
    begin
        if (G_ADDR_WIDTH = 4) then
            --! Deoxys rounds
            return 14;
        else
            --! Joltik rounds
            return 32;
        end if;
    end function get_round;
            
    constant MSB_AD                 : std_logic_vector(3 downto 0) := "0010";
    constant MSB_AD_LAST            : std_logic_vector(3 downto 0) := "0110";
    constant MSB_M                  : std_logic_vector(3 downto 0) := "0000";
    constant MSB_M_LAST_NONZERO     : std_logic_vector(3 downto 0) := "0100";
    constant MSB_CHKSUM_FULL        : std_logic_vector(3 downto 0) := "0001";
    constant MSB_CHKSUM_NON_FULL    : std_logic_vector(3 downto 0) := "0101";
    constant ROUNDS                 : integer                      := get_round;

    --! State
    type state_type is (
        S_INIT,             S_LD_KEY,   S_LD_NPUB,
        S_WAIT_BDI,
        S_PREP_TWEAKEY,     S_PROCESS);
    signal state, nstate    : state_type;
    --! Internal counters
    signal addr_s           : std_logic_vector(G_ADDR_WIDTH         -1 downto 0);
    signal count            : std_logic_vector(G_ADDR_WIDTH         -1 downto 0);
    --! Msg status register
    signal op_decrypt       : std_logic;
    --! Block status registers
    signal is_eot           : std_logic;
    signal is_eoi           : std_logic;
    signal is_tag           : std_logic;
    signal is_ad            : std_logic;
    signal is_init          : std_logic;
    signal is_full       : std_logic;
    --! Block status controls
    signal ld_blk_status    : std_logic;
    signal clr_blk_status   : std_logic;
    signal set_decrypt      : std_logic;
    signal clr_decrypt      : std_logic;
    signal sel_decrypt_s    : std_logic;
    --! Status translation
    signal bdi_ad           : std_logic;
    --! Tag block indicator controls
    signal clr_tag          : std_logic;
    signal set_tag          : std_logic;
    --! Tag block completion
    signal is_tag_done      : std_logic;
    signal set_tag_done     : std_logic;
    signal clr_tag_done     : std_logic;
    --! Addr controls
    signal ld_addr          : std_logic;
    signal en_addr          : std_logic;
    --! Init controls
    signal set_init         : std_logic;
    signal clr_init         : std_logic;
begin
    p_reg:
    process( clk )
    begin
        if rising_edge( clk ) then
            if rst = '1' then
                state      <= S_INIT;
                is_full <= '0';
            else
                state <= nstate;

                if (ld_addr = '1') then
                    count <= (others => '0');
                    if (sel_decrypt_s = '1' and state = S_PREP_TWEAKEY) then
                        addr_s <= std_logic_vector(to_unsigned(ROUNDS-2,G_ADDR_WIDTH));
                    else
                        addr_s <= (others => '0');
                    end if;
                elsif (en_addr = '1') then
                    count <= std_logic_vector(unsigned(count) + 1);
                    if (sel_decrypt_s = '1' and state = S_PROCESS) then
                        addr_s <= std_logic_vector(unsigned(addr_s) - 1);
                    else
                        addr_s <= std_logic_vector(unsigned(addr_s) + 1);
                    end if;
                end if;


                if (state = S_LD_NPUB) then
                    op_decrypt <= decrypt;
                end if;

                if (ld_blk_status = '1') then
                    is_eot <= bdi_eot;
                    is_eoi <= bdi_eoi;
                    is_ad  <= bdi_ad;
                    is_full <= bdi_size(G_LBS_BYTES);
                elsif (clr_blk_status = '1') then
                    is_eot <= bdi_eot;
                    is_eoi <= bdi_eoi;
                    is_ad  <= bdi_ad;
                end if;

                if (set_decrypt = '1') then
                    sel_decrypt_s <= '1';
                elsif (clr_decrypt = '1') then
                    sel_decrypt_s <= '0';
                end if;

                if (set_tag = '1') then
                    is_tag <= '1';
                elsif (clr_tag = '1') then
                    is_tag <= '0';
                end if;

                if (set_init = '1') then
                    is_init <= '1';
                elsif (clr_init = '1') then
                    is_init <= '0';
                end if;

                if (set_tag_done = '1') then
                    is_tag_done <= '1';
                elsif (clr_tag_done = '1') then
                    is_tag_done <= '0';
                end if;
            end if;
        end if;
    end process;
    addr        <= addr_s;
    sel_decrypt <= sel_decrypt_s;
    sel_chksum  <= op_decrypt;
    sel_output  <= is_full;
    sel_tag     <= is_tag_done;
    bdi_ad      <= '1' when bdi_type(2 downto 1) = "00" 
                            or bdi_type = BDI_TYPE_NPUB 
                       else '0';

    p_ss:
    process(is_tag, bdi_ad, bdi_eot, bdi_size, is_ad, is_full)
    begin
        if (is_tag = '0') then
            if (bdi_ad = '1') then
                if (bdi_eot = '1' and not bdi_size(G_LBS_BYTES) = '1') then
                    stage_status <= MSB_AD_LAST;
                else
                    stage_status <= MSB_AD;
                end if;
                sel_npub <= '0';
            else
                if (bdi_eot = '1' and not bdi_size(G_LBS_BYTES) = '1') then
                   stage_status <= MSB_M_LAST_NONZERO;
                else
                    stage_status <= MSB_M;
                end if;
                sel_npub <= '1';
            end if;
        else
            sel_npub <= '1';
            if (is_ad = '1' or is_full = '1') then
                stage_status <= MSB_CHKSUM_FULL;
            else
                stage_status <= MSB_CHKSUM_NON_FULL;
            end if;
        end if;
    end process;


    p_state:
    process( state, key_valid, key_update,
        bdi_valid, bdi_ad, bdi_eoi, bdi_eot, count,
        is_eot, is_eoi,
        is_full, is_tag, is_tag_done, is_init, sel_decrypt_s, op_decrypt,
        bdo_ready)
    begin
        --! Default values
        nstate          <= state;
        sel_input       <= "00";
        en_key          <= '0';
        en_block        <= '0';
        en_tweakey      <= '0';
        en_chksum       <= '0';
        ld_block        <= '0';
        ld_tweakey      <= '0';
        ld_chksum       <= '0';
        ld_auth         <= '0';
        en_auth         <= '0';
        en_npub         <= '0';
        bdo_valid       <= '0';
        bdi_ready       <= '0';
        key_ready       <= '0';
        ld_addr         <= '0';
        en_addr         <= '0';
        set_decrypt     <= '0';
        clr_decrypt     <= '0';
        ld_blk_status   <= '0';
        clr_blk_status  <= '0';
        set_tag         <= '0';
        clr_tag         <= '0';
        set_tag_done    <= '0';
        clr_tag_done    <= '0';
        set_init        <= '0';
        clr_init        <= '0';
        ld_state        <= '0';
        en_state        <= '0';
        msg_auth_done   <= '0';
        sel_rdkey       <= '0';

        case state is
            when S_INIT =>
                ld_addr         <= '1';
                ld_chksum       <= '1';
                ld_auth         <= '1';
                ld_block        <= '1';
                clr_blk_status  <= '1';
                clr_tag         <= '1';
                clr_tag_done    <= '1';
                set_init        <= '1';            
                if (bdi_valid = '1') then
                    if (key_update = '1') then
                        nstate <= S_LD_KEY;
                    else
                        nstate <= S_LD_NPUB;
                    end if;
                end if;

            when S_LD_KEY =>
                key_ready <= '1';
                if (key_valid = '1') then
                    en_key  <= '1';
                    en_addr <= '1';
                    if (unsigned(count) = 3) then
                        nstate  <= S_LD_NPUB;
                    end if;
                end if;

            when S_LD_NPUB =>
                ld_addr   <= '1';
                bdi_ready <= '1';
                if (bdi_valid = '1') then
                    en_npub <= '1';
                    nstate <= S_WAIT_BDI;
                    if (is_eoi = '1' and is_tag = '0') then
                        set_tag <= '1';
                    end if;
                end if;

            when S_WAIT_BDI =>
                --! Block number
                if (((bdi_valid = '1' or is_tag = '1')
                        and (is_init = '1'
                            or (is_init = '0' 
                                and (bdo_ready = '1' or is_ad = '1'))))
                    or (is_tag_done = '1'
                        and ((op_decrypt = '1' and bdi_valid = '1')
                            or (op_decrypt = '0' and bdo_ready = '1'))))
                then
                    if (is_tag_done = '0') then
                        if (is_ad = '1' and is_init = '0') then
                            en_auth   <= '1';
                        end if;
                        if ((op_decrypt = '0' and bdi_ad = '0'
                                and is_tag = '0')
                            or (op_decrypt = '1' and is_init = '0'
                                and is_ad = '0'))
                        then
                            en_chksum <= '1';
                        end if;
                    end if;
                    if (is_init = '0' and is_ad = '0'
                        and is_tag_done = '0')
                    then
                        bdo_valid <= '1';
                    end if;

                    ld_tweakey  <= '1';
                    ld_addr     <= '1';
                    if (bdi_ad = '0' and op_decrypt = '1'
                        and (bdi_eot = '0' or bdi_size(G_LBS_BYTES) = '1')
                        and is_tag = '0')
                    then
                        set_decrypt <= '1';
                    else
                        clr_decrypt <= '1';
                    end if;

                    if (is_tag_done = '1') then
                        if (op_decrypt = '0') then
                            nstate          <= S_INIT;
                            bdo_valid       <= '1';
                        elsif (op_decrypt = '1') then
                            nstate          <= S_INIT;
                            bdi_ready       <= '1';
                            msg_auth_done   <= '1';
                        end if;
                    else
                        nstate      <= S_PREP_TWEAKEY;
                    end if;
                end if;

            when S_PREP_TWEAKEY =>
                en_addr     <= '1';
                clr_init    <= '1';
                en_tweakey  <= '1';
                if (unsigned(count) = ROUNDS-1) then
                    nstate   <= S_PROCESS;
                    ld_addr  <= '1';
                    if (is_tag = '0') then
                        bdi_ready     <= '1';
                        ld_blk_status <= '1';
                    end if;
                    ld_state        <= '1';
                    if (is_tag = '1') then
                        sel_input <= "10";
                    elsif (bdi_ad = '0' and bdi_eot = '1'
                          and not bdi_size(G_LBS_BYTES) = '1')
                    then
                        sel_input <= "01";
                    end if;
                end if;

            when S_PROCESS =>
                en_state <= '1';
                en_addr  <= '1';
                if (unsigned(count) = ROUNDS-1) then
                    nstate  <= S_WAIT_BDI;
                    if (is_eoi = '1' and is_tag = '0') then
                        set_tag <= '1';
                    end if;
                    if (is_tag = '1') then
                        set_tag_done <= '1';
                    end if;
                    if (is_eot = '1' and is_eoi = '0') then
                        ld_block <= '1';
                    end if;
                    if (is_full = '1') then
                        en_block <= '1';
                    end if;
                    if (sel_decrypt_s = '1') then
                        sel_rdkey <= '1';
                    end if;
                end if;

        end case;
    end process;

end behavior;

